@@ -1,6 +1,8 @@
 # -*- coding: utf-8 -*-
 """GitHub class file."""
 
+import datetime
+# import logging
 import requests
 
 
@@ -41,6 +43,11 @@ class GitHub(object):
     #
     # Sub Classes
     #
+    def client(self, auth):
+        """Return a Client instance."""
+        from bits.github.client import Client
+        return Client(auth, self)
+
     def datastore(self, auth):
         """Return a Datastore instance."""
         from bits.github.datastore import Datastore
@@ -51,6 +58,11 @@ class GitHub(object):
         from bits.github.firestore import Firestore
         return Firestore(auth, self)
 
+    def sync(self, auth):
+        """Return an Update instance."""
+        from bits.github.sync import Sync
+        return Sync(auth, self)
+
     def update(self, auth):
         """Return an Update instance."""
         from bits.github.update import Update
@@ -73,6 +85,31 @@ class GitHub(object):
         # return json
         return response
 
+    def get_dict(self, items, key='id'):
+        """Return a list as a dict with the given key."""
+        data = {}
+        for item in items:
+            k = item[key]
+            data[k] = item
+        return data
+
+    def get_modified(self, url, headers={}, params={}, etag=None, last_modified=None):
+        """Return a response from a GET call if modified, otherwise return False."""
+        # add etag and last-modified request headers
+        if etag:
+            headers['If-None-Match'] = etag
+        elif last_modified:
+            headers['If-Modified-Since'] = last_modified
+
+        # get response
+        response = self.get(url, headers=headers)
+
+        # check if not changed
+        if response.status_code == 304:
+            return False
+
+        return response.json()
+
     def get_list(self, base_url, url, headers={}, params={}):
         """Return a paginated list from a GET call."""
         items_list = []
@@ -82,6 +119,10 @@ class GitHub(object):
 
         next_url = '{}/{}'.format(base_url, url)
 
+        # set page size
+        if 'per_page' not in params:
+            params['per_page'] = 100
+
         while next_url:
             # get response to request
             response = requests.get(next_url, headers=headers, params=params)
@@ -116,6 +157,10 @@ class GitHub(object):
         # GET /orgs/:org/hooks
         return self.get_list(self.org_base_url, 'hooks')
 
+    def get_org_hooks_dict(self, key='id'):
+        """Return a dict of org hooks."""
+        return self.get_dict(self.get_org_hooks(), key)
+
     #
     # Org Invitations
     #
@@ -125,6 +170,10 @@ class GitHub(object):
         headers = {'Accept': 'application/vnd.github.dazzler-preview'}
         return self.get_list(self.org_base_url, 'invitations', headers=headers)
 
+    def get_org_invitations_dict(self, key='id'):
+        """Return a dict of org invitations."""
+        return self.get_dict(self.get_org_invitations(), key)
+
     def get_org_invitation_teams(self, invitation_id):
         """Return a list of organization invitation teams."""
         # GET /orgs/:org/invitations/:invitation_id/teams
@@ -142,10 +191,15 @@ class GitHub(object):
             'filter': filterString,
             'role': role,
         }
+        # add 2fa_disabled tag if insecure is true
         if insecure:
             params['filter'] = '2fa_disabled'
         return self.get_list(self.org_base_url, 'members', params=params)
 
+    def get_org_members_dict(self, filterString=None, role=None, insecure=False, key='id'):
+        """Return a dict of organization teams."""
+        return self.get_dict(self.get_org_members(filterString, role, insecure), key)
+
     #
     # Org Memberships
     #
@@ -171,7 +225,7 @@ class GitHub(object):
     # Org Outside Collaborators
     #
     def get_org_outside_collaborators(self, filterString=None, insecure=False):
-        """Return a list of organization outside_collaborators."""
+        """Return a list of organization outside collaborators."""
         # GET /orgs/:org/outside_collaborators
         headers = {
             # 'Accept': 'application/vnd.github.korra-preview'
@@ -183,6 +237,10 @@ class GitHub(object):
             params['filter'] = '2fa_disabled'
         return self.get_list(self.org_base_url, 'outside_collaborators', headers=headers, params=params)
 
+    def get_org_outside_collaborators_dict(self, filterString=None, insecure=False, key='id'):
+        """Return a dict of organization outside collaborators."""
+        return self.get_dict(self.get_org_members(filterString, insecure), key)
+
     #
     # Org Public Members
     #
@@ -197,6 +255,10 @@ class GitHub(object):
         # GET /orgs/:org/public_members
         return self.get_list(self.org_base_url, 'public_members')
 
+    def get_org_public_members_dict(self, key='id'):
+        """Return a dict of org public members."""
+        return self.get_dict(self.get_org_public_members(), key)
+
     #
     # Org Repos
     #
@@ -206,6 +268,10 @@ class GitHub(object):
         headers = {'Accept': 'application/vnd.github.baptiste-preview+json'}
         return self.get_list(self.org_base_url, 'repos', headers=headers)
 
+    def get_org_repos_dict(self, key='id'):
+        """Return a dict of org repos."""
+        return self.get_dict(self.get_org_repos(), key)
+
     #
     # Org Teams
     #
@@ -216,6 +282,10 @@ class GitHub(object):
         headers = {'Accept': 'application/vnd.github.hellcat-preview+json'}
         return self.get_list(self.org_base_url, 'teams', headers)
 
+    def get_org_teams_dict(self, key='id'):
+        """Return a dict of organization teams."""
+        return self.get_dict(self.get_org_teams(), key)
+
     def get_org_team_hierarchy(self):
         """Return a dict of organization teams and their children."""
         organization = {}
@@ -234,18 +304,38 @@ class GitHub(object):
                     organization[pid]['children'].append(team)
         return organization
 
+    #
+    # Rate Limit
+    #
+    def get_rate_limit(self):
+        """Return a repo of the organization."""
+        # GET /rate_limit
+        url = '{}/rate_limit'.format(self.base_url)
+        data = requests.get(url, headers=self.headers).json()
+
+        # get rate information
+        rate = data['rate']
+        limit = rate['limit']
+        remaining = rate['remaining']
+        reset = rate['reset']
+
+        # calculate when the quota will reset
+        reset_time = datetime.datetime.fromtimestamp(reset)
+        now = datetime.datetime.now()
+        delta = reset_time - now
+        print('{} remaining of {} limit, resets in {} seconds at: {}'.format(
+            remaining, limit, delta.seconds, reset_time,
+        ))
+        return data
+
     #
     # Repos
     #
-    def get_repo(self, repo, etag=None):
+    def get_repo(self, repo, etag=None, last_modified=None):
         """Return a repo of the organization."""
         # GET /repos/:owner/:repo
         url = '{}/repos/{}/{}'.format(self.base_url, self.org, repo)
-        headers = self.headers.copy()
-        if etag:
-            headers['If-None-Match'] = etag
-            return self.get(url, headers=headers)
-        return self.get(url, headers=headers).json()
+        return self.get_modified(url, etag=etag, last_modified=last_modified)
 
     def get_repo_collaborators(self, repo, affiliation=None):
         """Return a list of repo hook."""
@@ -254,12 +344,20 @@ class GitHub(object):
         params = {'affiliation': affiliation}
         return self.get_list(self.base_url, url, params=params)
 
+    def get_repo_collaborators_dict(self, repo, affiliation=None, key='id'):
+        """Return a dict of org repo collaborators."""
+        return self.get_dict(self.get_repo_collaborators(repo, affiliation), key)
+
     def get_repo_hooks(self, repo):
         """Return a list of repo hook."""
         # GET /repos/:owner/:repo/hooks
         url = 'repos/{}/{}/hooks'.format(self.org, repo)
         return self.get_list(self.base_url, url)
 
+    def get_repo_hooks_dict(self, repo, key='id'):
+        """Return a dict of org hooks."""
+        return self.get_dict(self.get_repo_hooks(repo), key)
+
     def update_repo_hook(self, repo, hook_id, body):
         """Update a repo hook."""
         # PATCH /repos/:owner/:repo/hooks/:hook_id
@@ -269,15 +367,11 @@ class GitHub(object):
     #
     # Teams
     #
-    def get_team(self, team_id, etag=None):
+    def get_team(self, team_id, etag=None, last_modified=None):
         """Return a team of the organization."""
         # GET /teams/:team_id
         url = '{}/teams/{}'.format(self.base_url, team_id)
-        headers = self.headers.copy()
-        if etag:
-            headers['If-None-Match'] = etag
-            return self.get(url, headers=headers)
-        return self.get(url, headers=headers).json()
+        return self.get_modified(url, etag=etag, last_modified=last_modified)
 
     def get_team_invitations(self, team_id):
         """Return a list of team invitations."""
@@ -286,6 +380,10 @@ class GitHub(object):
         url = 'teams/{}/invitations'.format(team_id)
         return self.get_list(self.base_url, url, headers=headers)
 
+    def get_team_invitations_dict(self, team_id, key='id'):
+        """Return a dict of team invitations."""
+        return self.get_dict(self.get_team_invitations(team_id), key)
+
     def get_team_members(self, team_id, role='all'):
         """Return a list of team members."""
         # GET /teams/:team_id/members
@@ -294,14 +392,22 @@ class GitHub(object):
         url = 'teams/{}/members'.format(team_id)
         return self.get_list(self.base_url, url, params=params)
 
+    def get_team_members_dict(self, team_id, role='all', key='id'):
+        """Return a dict of team members."""
+        return self.get_dict(self.get_team_members(team_id, role), key)
+
     def get_team_members_with_children(self, team_id, role='all'):
-        """Return a list of team members."""
+        """Return a list of team members with children."""
         # GET /teams/:team_id/members
         headers = {'Accept': 'application/vnd.github.hellcat-preview+json'}
         params = {'role': role}
         url = 'teams/{}/members'.format(team_id)
         return self.get_list(self.base_url, url, headers=headers, params=params)
 
+    def get_team_members_with_children_dict(self, team_id, role='all', key='id'):
+        """Return a dict of team members with children."""
+        return self.get_dict(self.get_team_members_with_children(team_id, role), key)
+
     def invite_team_member(self, team, member):
         """Invite a member to the organization."""
         # PUT /teams/:id/memberships/:username
@@ -310,27 +416,31 @@ class GitHub(object):
         return requests.put(url, headers=self.headers, params=params)
 
     def remove_team_member(self, team, member):
-        """Invite a member to the organization."""
+        """Remove a member from the organization."""
         # DELETE /teams/:id/memberships/:username
         url = '{}/teams/{}/memberships/{}'.format(self.base_url, team, member)
         return requests.delete(url, headers=self.headers)
 
     def get_team_repos(self, team_id):
-        """Return a list of team members."""
+        """Return a list of team repos."""
         # GET /teams/:team_id/repos
         # Add header to get the Nested Teams API
         # headers = {'Accept': 'application/vnd.github.hellcat-preview+json'}
         url = 'teams/{}/repos'.format(team_id)
-        return self.get_list(self.base_url, url,)
+        return self.get_list(self.base_url, url)
+
+    def get_team_repos_dict(self, team_id, key='id'):
+        """Return a dict of team repos."""
+        return self.get_dict(self.get_team_repos(team_id), key)
 
     #
     # User (self - logged in user)
     #
-    def get_self(self):
+    def get_self(self, etag=None, last_modified=None):
         """Return the logged in user."""
         # GET /user
         url = '{}/user'.format(self.base_url)
-        return self.get(url).json()
+        return self.get_modified(url, etag=etag, last_modified=last_modified)
 
     def get_self_repos(self):
         """Return a list of user repos."""
@@ -338,6 +448,10 @@ class GitHub(object):
         url = 'user/repos'
         return self.get_list(self.base_url, url)
 
+    def get_self_repos_dict(self, key='id'):
+        """Return a dict of user repos."""
+        return self.get_dict(self.get_self_repos(), key)
+
     def get_self_teams(self):
         """Return a list of teams for the logged-in user."""
         # GET /user/teams - requires "user" or "repo" scope
@@ -345,21 +459,25 @@ class GitHub(object):
         url = 'user/teams'
         return self.get_list(self.base_url, url, headers=headers)
 
+    def get_self_teams_dict(self, key='id'):
+        """Return a dict of team teams."""
+        return self.get_dict(self.get_self_teams(), key)
+
     #
     # Users
     #
-    def get_user(self, login, etag=None):
+    def get_user(self, login, etag=None, last_modified=None):
         """Return a single user."""
         # GET /users/:username
         url = '{}/users/{}'.format(self.base_url, login)
-        headers = self.headers.copy()
-        if etag:
-            headers['If-None-Match'] = etag
-            return self.get(url, headers=headers)
-        return self.get(url, headers=headers).json()
+        return self.get_modified(url, etag=etag, last_modified=last_modified)
 
     def get_user_repos(self, login):
         """Return a list of public repositories for the specified user."""
         # GET /users/:username/repos
         url = 'users/{}/repos'.format(login)
         return self.get_list(self.base_url, url)
+
+    def get_user_repos_dict(self, login, key='id'):
+        """Return a dict of public repositories for the specified user."""
+        return self.get_dict(self.get_user_repos(login), key)
